__all__ = [
    "IDMode",
    "Event",
    "FeatureData",
    "GameSchema",
    "TableSchema",
]

from . import IDMode
from . import Event
from . import FeatureData
from ogd.core.schemas.games import GameSchema
from ogd.core.schemas.tables import TableSchema


__all__ = [
	"utils",
	"schemas",
    "generators",
	"interfaces",
	"managers",
	"requests"
]

from . import utils
from . import schemas
from . import generators
from . import interfaces
from . import managers
from . import requests



# import standard libraries
from datetime import datetime, timedelta
from typing import Callable, Final, List, Optional, Union
# import local files
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.detectors.DetectorEvent import DetectorEvent
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.schemas.Event import Event
from ogd.core.schemas.ExtractionMode import ExtractionMode
from ogd.core.utils.typing import Map


class HintAndLeave(Detector):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    DEFAULT_THRESHOLD : Final[int] = 5


    def __init__(self, params:GeneratorParameters, trigger_callback:Callable[[Event], None], time_threshold:Optional[int]):
        super().__init__(params=params, trigger_callback=trigger_callback)
        self._found = False
        self._sess_id             : str = "Unknown"
        self._scene               : str = "Unknown"
        self._job_name            : str = "Unknown"
        self._hint                : str = "Unknown"
        self._last_hint_time      : Optional[datetime] = None
        self._time_spent          : Optional[Union[timedelta, float]] = None
        self._detector_event_data : Map = {}
        self._threshold           : timedelta
        if time_threshold is not None:
            self._threshold = timedelta(seconds=time_threshold)
        else:
            self._threshold = timedelta(seconds=HintAndLeave.DEFAULT_THRESHOLD)

    # *** Implement abstract functions ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["ask_for_help", "scene_changed", "room_changed"] # >>> fill in names of events this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        if self._sess_id == "Unknown":
            self._sess_id = event.SessionID
        elif self._sess_id != event.SessionID:
            self._sess_id = event.SessionID
            self._last_hint_time = None

        if event.EventName == "ask_for_help":
            self._last_hint_time = event.Timestamp
            self._job_name = event.GameState.get('job_name', event.EventData.get('job_name', None))
            self._hint = event.EventData.get("node_id", "NODE NOT FOUND")
            return

        if not self._last_hint_time:
            return
        self._time_spent = event.Timestamp - self._last_hint_time
        self._last_hint_time = None

        if self._time_spent <= self._threshold:
            self._found = True
            self._time_spent = self._time_spent / timedelta(seconds=1)
            self._sess_id = event.SessionID
            self._sequence_index = event.EventSequenceIndex
            if event.EventName == "scene_changed":
                self._scene = event.EventData.get("scene_name", "SCENE NOT FOUND")
                self._detector_event_data = {
                    "time": self._time_spent,
                    "level": self._threshold / timedelta(seconds=1),
                    "scene": self._scene,
                    "job_name": self._job_name,
                    "hint_node": self._hint
                }
            else:
                self._room = event.EventData.get("room_name")
                self._detector_event_data = {
                    "time": self._time_spent,
                    "level": self._threshold / timedelta(seconds=1),
                    "room": self._room,
                    "job_name": self._job_name,
                    "hint_node": self._hint
                }
        return

    def _trigger_condition(self) -> bool:
        if self._found:
            self._found = False
            return True
        else:
            return False

    def _trigger_event(self) -> DetectorEvent:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ret_val: DetectorEvent = self.GenerateEvent(session_id=self._sess_id, app_id="AQUALAB",
                                                    event_name="HintAndLeave", event_data=self._detector_event_data,
                                                    event_sequence_index=self._sequence_index)
        return ret_val



# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.PENGUINS.features.PerRegionFeature import PerRegionFeature
from ogd.core.schemas.Event import Event
from ogd.core.schemas.ExtractionMode import ExtractionMode
from ogd.core.schemas.FeatureData import FeatureData
from ogd.core.generators.extractors.SessionFeature import SessionFeature

# gaze_dict = { 'BigRock00': 0, 'Bridge': 0, 'Bridge2': 0, 'BigRock01': 0, 'River3': 0, 'Inland2': 0, 'Sea4': 0, 'River2': 0,'Inland1': 0, 'Sea5': 0, 'Sea3': 0, 'River5': 0, 'River4': 0, 'Inland5': 0, 'Inland4': 0, 'Inland3': 0,'River1': 0, 'Sea2': 0, 'Sea1': 0}
class GazeCount(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        # self._current_count : int = 0
        self._object_id = None
        self._gaze_dict = dict()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["gaze_object_end"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        # self._current_count += 1
        
        self._object_id = event.event_data.get("object_id")
        if not self._object_id in self._gaze_dict.keys():
                self._gaze_dict[self._object_id] = 0
        
        self._gaze_dict[self._object_id]+=1
        
    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._gaze_dict]



from typing import Any, List

from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.schemas.Event import Event
from ogd.core.schemas.ExtractionMode import ExtractionMode
from ogd.core.schemas.FeatureData import FeatureData

class AppVersions(SessionFeature):
    """_summary_

    :param SessionFeature: _description_
    :type SessionFeature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._versions = set()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        self._versions.add(event.AppVersion)

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [list(self._versions)]

    # *** Optionally override public functions. ***


__all__ = [
    "SessionDuration",
    "RegionsEncountered",
    "RegionEnterCount",
    "PlayerWaddleCount",
    "GazeDuration",
    "GazeCount",
    "SnowBallDuration",
    # "RingChimesCount",
    "WaddlePerRegion",
    "EatFishCount",
    "PickupRockCount",
    "EggLostCount",
    "EggRecoverTime",
    "PlayerInactiveAvgDuration",
    "MirrorWaddleDuration",
    "WaddlePerRegion",
    "RegionDuration",
    "ActivityCompleted",
    "ActivityDuration"
]
# aggregated features

from . import SessionDuration
from . import RegionsEncountered
from . import PlayerWaddleCount
from . import GazeDuration
from . import GazeCount
from . import SnowBallDuration
from . import RingChimesCount
from . import EatFishCount
from . import PickupRockCount
from . import EggLostCount
from . import EggRecoverTime
from . import PlayerInactiveAvgDuration
from . import MirrorWaddleDuration
# percount features
from . import RegionEnterCount
from . import WaddlePerRegion
from . import RegionDuration
from . import ActivityCompleted
from . import ActivityDuration




# import libraries
from ogd.core.schemas import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.schemas.Event import Event
from ogd.core.schemas.ExtractionMode import ExtractionMode
from ogd.core.schemas.FeatureData import FeatureData

class PersistentSessionID(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        SessionFeature.__init__(self, params=params)
        self._persistent_id : Optional[int] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["BEGIN.0"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if self._persistent_id is None:
            self._persistent_id = event.UserData['persistent_session_id']

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._persistent_id]

    # *** Optionally override public functions. ***





# import standard libraries
import logging
import re
from typing import Any, Dict, Optional, List

# import 3rd-party libraries

# import OGD libraries
from ogd.core.utils.Logger import Logger

# import locals

class SemanticVersion:
    def __init__(self, major:Optional[int], minor:Optional[int]=None, fix:Optional[int]=None, suffix:Optional[str]=None, suffix_ver:Optional[int]=None, fallback:Optional[str]=None):
        self._major      : Optional[int] = major
        self._minor      : Optional[int] = minor
        self._fix        : Optional[int] = fix
        self._suffix     : Optional[str] = suffix
        self._suffix_ver : Optional[int] = suffix_ver
        self._fallback   : Optional[str] = fallback

    def __repr__(self) -> str:
        return f"SemanticVersion : {self}"

    def __str__(self) -> str:
        _major      = str(self._major)       if self._major      is not None else None
        _minor      = f".{self._minor}"      if self._minor      is not None else None
        _fix        = f".{self._fix}"        if self._fix        is not None else None
        _suffix     = f"-{self._suffix}"     if self._suffix     is not None else None
        _suffix_ver = f".{self._suffix_ver}" if self._suffix_ver is not None else None

        return f"{_major}{_minor or ''}{_fix or ''}{_suffix or ''}{_suffix_ver or ''}" if _major is not None else str(self._fallback)

    def __eq__(self, other) -> bool:
        if type(other) != type(self):
            if isinstance(other, str):
                return self == SemanticVersion.FromString(other, verbose=False)
            return False
        if self._major is None or other._major is None:
            return False
        if self._major       == other._major  \
        and self._minor      == other._minor  \
        and self._fix        == other._fix    \
        and self._suffix     == other._suffix \
        and self._suffix_ver == other._suffix_ver:
            return True
        return False


    @staticmethod
    def FromString(semver:str, verbose:bool=True) -> 'SemanticVersion':
        pieces = re.split('\.|-', semver)
        if verbose:
            Logger.Log(f"Pieces: {pieces}", logging.DEBUG)

        return SemanticVersion._parseMajor(semver=semver, pieces=pieces, verbose=verbose)

    @staticmethod
    def _parseMajor(semver:str, pieces:List[str], verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _major : int
        try:
            _major = int(pieces[0])
        except ValueError as err:
            _default_ver = semver
            if verbose:
                Logger.Log(f"Could not parse semantic version from {semver}, {pieces[0]} was not an int! Defaulting to version={_default_ver}\nError Details: {err}", level=logging.WARN)
            ret_val = SemanticVersion(major=None, fallback=semver)
        else:
            if len(pieces) > 1:
                ret_val = SemanticVersion._parseMinor(semver=semver, pieces=pieces[1:], major=_major, verbose=verbose)
            else:
                ret_val = SemanticVersion(major=_major)
        finally:
            return ret_val
                

    @staticmethod
    def _parseMinor(semver:str, pieces:List[str], major:int, verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _minor : int
        try:
            _minor = int(pieces[0])
        except ValueError as err:
            _default_ver = str(major)
            if verbose:
                Logger.Log(f"Could not parse semantic version from {semver}, {pieces[0]} was not an int! Defaulting to version={_default_ver}\nError Details: {err}", level=logging.WARN)
            ret_val = SemanticVersion(major=major)
        else:
            if len(pieces) > 1:
                ret_val = SemanticVersion._parseFix(semver=semver, pieces=pieces[1:], major=major, minor=_minor, verbose=verbose)
            else:
                ret_val = SemanticVersion(major=major, minor=_minor)
        finally:
            return ret_val

    @staticmethod
    def _parseFix(semver:str, pieces:List[str], major:int, minor:int, verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _fix : int
        try:
            _fix = int(pieces[0])
        except ValueError as err:
            _default_ver = f"{major}.{minor}"
            if verbose:
                Logger.Log(f"Could not parse semantic version from {semver}, {pieces[0]} was not an int! Defaulting to version={_default_ver}\nError Details: {err}", level=logging.WARN)
            ret_val = SemanticVersion(major=major, minor=minor)
        else:
            if len(pieces) > 1:
                ret_val = SemanticVersion._parseSuffix(semver=semver, pieces=pieces[1:], major=major, minor=minor, fix=_fix, verbose=verbose)
            else:
                ret_val = SemanticVersion(major=major, minor=minor, fix=_fix)
        finally:
            return ret_val

    @staticmethod
    def _parseSuffix(semver:str, pieces:List[str], major:int, minor:int, fix:int, verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _suffix : str
        _suffix = pieces[0]
        if len(pieces) > 1:
            ret_val = SemanticVersion._parseSuffixVersion(semver=semver, pieces=pieces[1:], major=major, minor=minor, fix=fix, suffix=_suffix, verbose=verbose)
        else:
            ret_val = SemanticVersion(major=major, minor=minor, fix=fix, suffix=_suffix)
        return ret_val

    @staticmethod
    def _parseSuffixVersion(semver:str, pieces:List[str], major:int, minor:int, fix:int, suffix:str, verbose:bool) -> 'SemanticVersion':
        ret_val : 'SemanticVersion'

        _suffix_version : int
        try:
            _suffix_version = int(pieces[0])
        except ValueError as err:
            _default_ver = f"{major}.{minor}.{fix}-{suffix}"
            if verbose:
                Logger.Log(f"Could not parse semantic version from {semver}, {pieces[0]} was not an int! Defaulting to version={_default_ver}\nError Details: {err}", level=logging.WARN)
            ret_val = SemanticVersion(major=major, minor=minor, fix=fix, suffix=suffix)
        else:
            ret_val = SemanticVersion(major=major, minor=minor, fix=fix, suffix=suffix, suffix_ver=_suffix_version)
            if len(pieces) > 1 and verbose:
                Logger.Log(f"Semantic version string {semver} had excess parts, reducing to version={ret_val}", level=logging.WARN)
        finally:
            return ret_val


# import libraries
import logging
import json 
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.schemas.Event import Event



class PerRegionFeature(PerCountFeature):
    def __init__(self, params:GeneratorParameters, region_map:List[Dict[str, Any]]):
        super().__init__(params=params)
        self._region_map = region_map

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _validateEventCountIndex(self, event:Event):
        ret_val : bool = False

        # region_data = event.EventData["region_name"]
        target_region = self._region_map[self.CountIndex]
        current_position = {
            'x': event.GameState['posX'],
            'y': event.GameState['posY'],
            'z': event.GameState['posZ']
        }

        if (current_position['x'] > target_region['minX'] and 
            current_position['x'] < target_region['maxX'] and
            current_position['y'] > target_region['minY'] and
            current_position['y'] < target_region['maxY'] and
            current_position['z'] > target_region['minZ'] and
            current_position['z'] < target_region['maxZ']):
            ret_val = True
        # else:
        #     self.WarningMessage(f"Got invalid region_name data in {type(self).__name__}")
        return ret_val

        #if region_data is not None:
            #if region_data in self.region_map and self.region_map[region_data] == self.CountIndex:
                #ret_val = True
        #else:
            #self.WarningMessage(f"Got invalid region_name data in {type(self).__name__}")
        #return ret_val

    # *** Optionally override public functions. ***

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"
